/* 
 *  This program is an interactive FAQ to ease investigations.
 *  Copyright (C) 2012 Julien Bruguier.
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 * 
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */ 

#ifndef __INVESTIGATION_DONNEES_COMPTEURS__
#define __INVESTIGATION_DONNEES_COMPTEURS__

#include <iomanip>
#include <map>
#include <src/global/global.h>
namespace Investigation { namespace Donnees { DECL_SHARED_PTR(Compteurs); } }
#include <src/investigation/donnees/statistiques.h>
#include <src/investigation/donnees/identifiants.h>
#include <src/investigation/donnees/selectionpage.h>
#include <src/investigation/donnees/selectionreponse.h>

namespace Investigation
{
	namespace Donnees
	{
		DECL_SHARED_PTR(Condition);
		class Site;
		class ArbreInvestigation;

		struct LienPage
		{
			LienPage(const Identifiant& page_depart, const std::string& mot_clef)
			:_page_depart(page_depart), _mot_clef(mot_clef) {};
			bool operator<(const LienPage& lien) const
			{
				if(_page_depart<lien._page_depart)
					return true;
				if(_page_depart>lien._page_depart)
					return false;
				return _mot_clef<lien._mot_clef;
			}
			Identifiant _page_depart;
			std::string _mot_clef;
		};

		class Compteurs
		{
			public:
				Compteurs()
					:_vrai(0), _faux(0), _non_evalue(0) {}
				void operator+=(const Statistiques::Resultat& resultat);
				operator bool() const
				{
					return (_vrai!=0 and _faux!=0);
				}
				template<typename oStream>
				friend oStream& operator<<(oStream& os, const Compteurs& compteurs)
				{
#ifdef INVESTIGATION_DEBUG
					os	<< "True: " << compteurs._vrai << "   False: " << compteurs._faux <<  "   N/A: " << compteurs._non_evalue;
#else
					if(compteurs._vrai==0)
					{
						if(compteurs._faux==0)
						{
							os << "Never evaluated";
						}
						else
						{
							os << "Always false   ";
						}
					}
					else
					{
						if(compteurs._faux==0)
						{
							os << "Always true    ";
						}
						else
						{
							os << "OK             ";
						}
					}
#endif
					return os;
				}
				size_t _vrai;
				size_t _faux;
				size_t _non_evalue;
			private:
		};

		struct Frequence
		{
			typedef enum class {NON_ATTEINT,EXCEPTIONNEL,RARE,NORMAL,FREQUENT,GENERIQUE} FrequencePage;
			static std::string frequence_page(const FrequencePage frequence)
			{
				switch(frequence)
				{
					case FrequencePage::NON_ATTEINT:
						return "Unreachable ";
					case FrequencePage::EXCEPTIONNEL:
						return "Exceptionnal";
					case FrequencePage::RARE:
						return "Rare        ";
					case FrequencePage::NORMAL:
						return "Normal      ";
					case FrequencePage::FREQUENT:
						return "Frequent    ";
					case FrequencePage::GENERIQUE:
						return "Generic     ";
					default:
						throw;
				}
			}
		};

		class CompteursSite
		{
			friend class ArbreInvestigation;
			public:
				struct CompteurConditionPage
				{
					CompteurConditionPage(const Investigation::Global::Source::PositionSP& position, const ConditionSP& condition)
						:_position(position), _condition(condition) { };
					Investigation::Global::Source::PositionSP _position;
					ConditionSP _condition;
					Compteurs _compteurs;
					template<typename oStream>
					friend oStream& operator<<(oStream& os, const CompteurConditionPage& conditionpage)
					{
						if(conditionpage._compteurs)
							return os;
						std::string position = *(conditionpage._position);
						std::string condition = *(conditionpage._condition);
						os	<< "    " << conditionpage._compteurs << "   " <<  position << condition << std::endl;
						return os;
					}
				};

				struct CompteurPage
				{
					CompteurPage(const Investigation::Global::Source::PositionSP& position, const std::set<std::string>& identifiants, const std::string& titre)
						:_position(position), _identifiants(identifiants), _titre(titre) { };
					Investigation::Global::Source::PositionSP _position;
					std::set<std::string> _identifiants;
					std::string _titre;
					Compteurs _compteurs;
					operator std::string() const
					{
						std::ostringstream oss;
						std::string position = *_position;
						oss << position << "Page (";
						bool en_cours = false;
						for(auto it=_identifiants.cbegin() ; it!=_identifiants.cend() ; ++it)
						{
							if(en_cours)
							{
								oss << ", ";
							}
							else
							{
								en_cours = true;
							}
							oss << *it;
						}
						oss << ") " << _titre ;
						return oss.str();
					}
					void ajout(std::map<size_t,size_t>& distribution) const
					{
						auto it=distribution.find(_compteurs._vrai);
						if(it==distribution.end())
						{
							distribution[_compteurs._vrai]=1;
							return;
						}
						++(it->second);
					}
					template<typename oStream>
					friend oStream& operator<<(oStream& os, const CompteurPage& page)
					{
						std::string titre = page;
						os << titre << " ===========> " << page._compteurs._vrai << " / " << page._compteurs._faux + page._compteurs._non_evalue;
						return os;
					}
					template<typename oStream>
					void affiche_frequence(const std::map<size_t,Frequence::FrequencePage>& notes_pages, const std::set<std::string>& mots_clef_reponses, oStream& os) const
					{
						std::string titre = *this;
						Frequence::FrequencePage frequence = notes_pages.at(_compteurs._vrai);
						std::string explication = "                 ";
						if(frequence==Frequence::FrequencePage::NON_ATTEINT)
						{
							bool atteint=false;
							for(auto it=_identifiants.cbegin() ; it!=_identifiants.cend() ; ++it)
							{
								if(mots_clef_reponses.find(*it)!=mots_clef_reponses.end())
								{
									atteint=true;
									break;
								}
							}
							if(atteint)
							{
								if(_compteurs._faux==0)
								{
									 explication = "     Never tested";
								}
								else
								{
									explication =  "      Too complex";
								}
							}
							else
							{
								explication =          "No answer to page";
							}
						}
						os << explication << " | " << Frequence::frequence_page(frequence) << " | " << titre;
#ifdef INVESTIGATION_DEBUG
						os << "   (" << _compteurs._vrai << ", " << _compteurs._faux << ", " << _compteurs._non_evalue << ")";
#endif
					}
				};

				struct CompteurConditionReponse
				{
					CompteurConditionReponse(const Investigation::Global::Source::PositionSP& position, const ConditionSP& condition)
						:_position(position), _condition(condition) { };
					Investigation::Global::Source::PositionSP _position;
					ConditionSP _condition;
					Compteurs _compteurs;
					template<typename oStream>
					friend oStream& operator<<(oStream& os, const CompteurConditionReponse& conditionreponse)
					{
						if(conditionreponse._compteurs)
							return os;
						std::string position = *(conditionreponse._position);
						std::string condition = *(conditionreponse._condition);
						os	<< "        " << conditionreponse._compteurs << "   " <<  position << condition << std::endl;
						return os;
					}
				};

				void operator+=(const Site& site);
				void operator+=(const Selection<Instruction>::Clause& conditionpage);
				void operator+=(const Page& page);
				void operator+=(const Selection<Reponse>::Clause& conditionreponse);
				void operator+=(const std::string& motclef);
				CompteursSite();
				template<typename oStream, int colonnes=8>
				static void categorie_page(const std::string& categorie, const size_t debut, const size_t fin, const size_t nb_pages, oStream& os) 
				{
					if(debut<=fin)
					{
						os 	<< "          " << categorie << ": " << std::setw(colonnes) << debut;
						if(debut!=fin)
						{
							os << " -" << std::setw(colonnes) << fin;
						}
						else
						{
							os << "  " << std::string(colonnes,' ');
						}
						os << " (" << std::setw(colonnes) << nb_pages << " page" << ((nb_pages==1)?" ":"s") << ")" << std::endl;
					}

				}
				template<typename oStream>
					friend oStream& operator<<(oStream& os, const CompteursSite& compteurs)
					{
						os	<< std::endl
							<< "--------------------------------------------------------------------------------------------------------------------------------------------------------"<< std::endl
							<< std::endl
							<< "Legend:" << std::endl
							<< "  Site completion is the proportion of non draft pages." << std::endl
							<< std::endl
							<< "  Pages profile is the distribution of pages categorized by the number of question and text items." << std::endl
							<< std::endl
							<< "  Identifiers sets are the different sets of identifiers detected as representing the same words." << std::endl
							<< "    Identifiers are considered equivalent when the sets of words separated by underscores are differents from one word at most," << std::endl
							<< "    each word is equivalent to another one when it is a prefix of this word, or have at most two different letters and have the" << std::endl
							<< "    common letters more or less in the same order," << std::endl
							<< std::endl
							<< "  Conditions on pages and answers are conditions with potential issue detected during the basic investigation done by the tool." << std::endl
							<< std::endl
							<< "  Reverse links are links followed during the investigation but presented by destination page. For each destination page, the list of source pages and" << std::endl
							<< "    keyword associated to the link is displayed. For each source, the proportion of the link usage is displayed." << std::endl
							<< std::endl
							<< "  Pages distribution is a summary of the number of occurences of each page found during the basic investigation:" << std::endl
							<< "    Six levels are shown by the tool:" << std::endl
							<< "    - \"Exceptionnal\", \"Rare\", \"Normal\", \"Frequent\", \"Generic\" are level of reached pages, the occurences ranges are display before the list of pages," << std::endl
							<< "    - \"Unreachable\" is the level of non reached pages, with an explanation about a potential issue:" << std::endl
							<< "      . \"No answer to page\" indicates that the page can not be reached by design, as no answer can activate this page even when all conditions are true," << std::endl
							<< "      . \"Too complex\" indicates that answers can activate this page, but this situation has never been obtained by the tool, meaning that this page" << std::endl
							<< "        will never be seen during a simple investigation by a user," << std::endl
							<< "      . \"Never tested\" indicates that answers can activate this page, but the conditions around it has been false during the whole investigation." << std::endl
							<< std::endl
							<< "--------------------------------------------------------------------------------------------------------------------------------------------------------"<< std::endl
							<< std::endl
							<< "Site statistics:" << std::endl
							<< "  Page count       : " << compteurs._nombre_pages/compteurs._nombre_requetes << std::endl
							<< "  Site completion  : " << (100*(compteurs._nombre_pages-compteurs._nombre_pages_brouillon))/(compteurs._nombre_pages) << " %" << std::endl
							<< "  Pages with title : " << (100*compteurs._nombre_pages_avec_titre)/compteurs._nombre_pages  << " %" << std::endl;
						os << std::endl << std::endl;
						{
							os	<< "  Pages profile    : " << std::endl;
							auto itd = compteurs._distribution_niveau_information.cbegin()->first;
							if(itd>0)
								itd=0;
							auto itf = compteurs._distribution_niveau_information.crbegin()->first;
							if(itf<0)
								itf=0;
							for(auto it=itf ; it>=itd ; --it)
							{
								auto val=compteurs._distribution_niveau_information.find(it);
#ifdef INVESTIGATION_DEBUG
								os << "    " << it << " : ";
								if(val!=compteurs._distribution_niveau_information.end())
								{
									os << (100*val->second)/compteurs._nombre_pages << " %";
								}
								else
								{
									os << 0 <<" %";
								}
								os << std::endl;
#endif
								os << "                ";
								if((it==itd) and (itd!=0))
								{
									os << "More questions";
								}
								else if((it==itf) and (itf!=0))
								{
									os << "  More answers";
								}
								else if(it==0)
								{
									os << "      Balanced";
								}
								else
								{
									os << "              ";
								}
								os << " | ";
								if(val!=compteurs._distribution_niveau_information.end())
								{
									size_t valeur = (100*val->second)/compteurs._nombre_pages;
									for(size_t i=0 ; i<valeur ; i+=5)
									{
										os << "#";
									}
									switch(valeur%5)
									{
										case 0: case 1: 
											break;
										case 2:
											os << ":";
											break;
										case 3: case 4:
											os << "#";
											break;
										default:
											throw;
									}
								}
								os << std::endl;
							}
						}
						os << std::endl << std::endl;
						{
							os << "  Identifiers sets :" << std::endl;
							std::set<std::string> mots_clef = compteurs._mots_clef_pages;
							mots_clef.insert(compteurs._mots_clef_reponses.cbegin(),compteurs._mots_clef_reponses.cend());
							std::set<std::set<std::string> > identifiants_groupes = groupes_identifiants(mots_clef);
							for(auto it=identifiants_groupes.cbegin() ; it!=identifiants_groupes.cend() ; ++it)
							{
								if(it->size()>1)
								{
									os << "    -";
									for(auto itit=it->cbegin() ; itit!=it->cend() ; ++itit)
									{
										os << " " << *itit;
									}
									os << std::endl;
								}
							}
						}
						os << std::endl << std::endl;
						os << "Site investigation:" << std::endl;
						{
							os << "  Conditions on pages:" << std::endl;
							for(auto it=compteurs._conditions_page.cbegin() ; it!=compteurs._conditions_page.cend() ; ++it)
							{
								os << it->second;
							}
						}
						os << std::endl << std::endl;
						{
							os << "  Conditions on answers:" << std::endl;
							for(auto it=compteurs._conditions_reponse.cbegin() ; it!=compteurs._conditions_reponse.cend() ; ++it)
							{
								if(not it->second.empty())
								{
									std::string page = compteurs._pages.at(it->first);
									os << "    " << page << ":" << std::endl;
									for(auto itit=it->second.cbegin() ; itit!=it->second.cend() ; ++itit)
									{
										os << itit->second;
									}
									os << std::endl;
								}
							}
						}
						os << std::endl;
						{
							os << "  Reverse links:" << std::endl;
							std::ostringstream mots_clef_inutilises_oss;
							mots_clef_inutilises_oss << "    Unused identifiers:" << std::endl;
							for(auto it_dest=compteurs._liens_inverses.cbegin() ; it_dest!=compteurs._liens_inverses.cend() ; ++it_dest)
							{
								std::string page_destination = compteurs._pages.at(it_dest->first);
								os << "    " << page_destination << ":" << std::endl;
								size_t total = 0;
								for(auto it_source=it_dest->second.cbegin() ; it_source!=it_dest->second.cend() ; ++it_source)
								{
									total += it_source->second;
								}
								std::set<std::string> mots_clef_utilises;
								for(auto it_source=it_dest->second.cbegin() ; it_source!=it_dest->second.cend() ; ++it_source)
								{
									mots_clef_utilises.insert(it_source->first._mot_clef);
									std::string page_source = compteurs._pages.at(it_source->first._page_depart);
									os << "      " << std::setw(5) << std::fixed << std::setprecision(1) << (100.*it_source->second)/total << "% " << page_source << ": " << it_source->first._mot_clef << std::endl;
								}
								if(mots_clef_utilises.size()!=compteurs._pages.at(it_dest->first)._identifiants.size())
								{
									mots_clef_inutilises_oss << "      " << page_destination << ":";
									for(auto it=compteurs._pages.at(it_dest->first)._identifiants.cbegin() ; it!=compteurs._pages.at(it_dest->first)._identifiants.cend() ; ++it)
									{
										if(mots_clef_utilises.find(*it)==mots_clef_utilises.end())
										{
											mots_clef_inutilises_oss << " " << *it;
										}
									}
									mots_clef_inutilises_oss << std::endl;
								}
								os << std::endl;
							}
							os << mots_clef_inutilises_oss.str() << std::endl;
						}
						os << std::endl;
						{
							os << "  Pages distribution:" << std::endl;
							std::map<size_t,size_t> distribution;
							for(auto it=compteurs._pages.cbegin() ; it!=compteurs._pages.cend() ; ++it)
							{
#ifdef INVESTIGATION_DEBUG
								os << "    " << it->second << std::endl;
#endif
								it->second.ajout(distribution);
							}
							size_t total=0;
							for(auto it=distribution.cbegin() ; it!=distribution.cend() ; ++it)
							{
								if(it->first!=0)
								{
									total += it->second;
								}
							}
							size_t cumul=0;
							std::map<size_t,Frequence::FrequencePage> notes_pages;
							size_t exceptionnel=0;
							size_t rare=0;
							size_t normal=0;
							size_t frequent=0;
							size_t generique=0;
							size_t nb_pages_non_atteintes=0;
							size_t nb_pages_exceptionnel=0;
							size_t nb_pages_rare=0;
							size_t nb_pages_normal=0;
							size_t nb_pages_frequent=0;
							size_t nb_pages_generique=0;
							for(auto it=distribution.cbegin() ; it!=distribution.cend() ; ++it)
							{
								if(it->first!=0)
								{
									cumul+=it->second;
								}
								auto frequence = (100*cumul)/total;
								if(frequence>95)
								{
									notes_pages[it->first]=Frequence::FrequencePage::GENERIQUE;
									generique=it->first;
									nb_pages_generique += it->second;								}
								else if(frequence>75)
								{
									notes_pages[it->first]=Frequence::FrequencePage::FREQUENT;
									frequent=it->first;
									nb_pages_frequent += it->second;
								}
								else if(frequence>25)
								{
									notes_pages[it->first]=Frequence::FrequencePage::NORMAL;
									normal=it->first;
									nb_pages_normal += it->second;
								}
								else if(frequence>5)
								{
									notes_pages[it->first]=Frequence::FrequencePage::RARE;
									rare=it->first;
									nb_pages_rare += it->second;
								}
								else if(frequence>0)
								{
									notes_pages[it->first]=Frequence::FrequencePage::EXCEPTIONNEL;
									exceptionnel=it->first;
									nb_pages_exceptionnel += it->second;
								}
								else
								{
									notes_pages[it->first]=Frequence::FrequencePage::NON_ATTEINT;
									nb_pages_non_atteintes += it->second;
								}
							}
							os	<< "      Pages occurences distribution:" << std::endl;
							categorie_page(Frequence::frequence_page(Frequence::FrequencePage::NON_ATTEINT)  , 0             ,            0, nb_pages_non_atteintes, os);
							categorie_page(Frequence::frequence_page(Frequence::FrequencePage::EXCEPTIONNEL) , 0           +1, exceptionnel, nb_pages_exceptionnel , os);
							categorie_page(Frequence::frequence_page(Frequence::FrequencePage::RARE)         , exceptionnel+1,         rare, nb_pages_rare         , os);
							categorie_page(Frequence::frequence_page(Frequence::FrequencePage::NORMAL)       , rare        +1,       normal, nb_pages_normal       , os);
							categorie_page(Frequence::frequence_page(Frequence::FrequencePage::FREQUENT)     , normal      +1,     frequent, nb_pages_frequent     , os);
							categorie_page(Frequence::frequence_page(Frequence::FrequencePage::GENERIQUE)    , frequent    +1,    generique, nb_pages_generique    , os);
							os	<< std::endl;
							for(auto it=compteurs._pages.cbegin() ; it!=compteurs._pages.cend() ; ++it)
							{
								os << "    ";
								it->second.affiche_frequence(notes_pages,compteurs._mots_clef_reponses,os);
								os << std::endl;
							}
						}
						return os;
					};
			private:
				std::map<Identifiant,CompteurConditionPage> _conditions_page;
				std::map<Identifiant,CompteurPage> _pages;
				std::map<Identifiant,std::map<Identifiant,CompteurConditionReponse> > _conditions_reponse; // Id page puis Id condition
				IdentifiantSP _page_courante;
				size_t _nombre_requetes;
				size_t _nombre_pages;
				size_t _nombre_pages_brouillon;
				size_t _nombre_pages_avec_titre;
				std::map<long int,size_t> _distribution_niveau_information;
				std::set<std::string> _mots_clef_pages;
				std::set<std::string> _mots_clef_reponses;
				std::map<Identifiant,std::map<LienPage,size_t> > _liens_inverses;
				static std::set<std::set<std::string> > groupes_identifiants(const std::set<std::string>& identifiants);
		};

	}
}
#endif
